import re

from types import ModuleType
from typing import List, Optional, Type

from pydoctest.parsers.parser import Parameter, Parser
from pydoctest.utilities import get_type_from_module
from pydoctest.exceptions import ParseException

"""
Based on this style guide:
https://numpydoc.readthedocs.io/en/latest/format.html

It doesn't currently mention a version, but there's an issue on that now:
https://github.com/numpy/numpydoc/issues/333

"""


class NumpyParser(Parser):
    def __init__(self) -> None:
        """Parser for Numpy docstring style."""
        super().__init__()
        self.section_headers = [
            # Rest are not supported (Yields, See Also, Warns, Warnings, Receives)
            'Parameters', 'Returns', 'Raises',
        ]
        # Split by word, newline, a number of '-' and newline
        self.section_regex = re.compile("([a-zA-Z]+)\n[-]+\n")
        self.parameter_regex = re.compile(r"(\w+)\s*:\s*([\w\[\], ]+)")
        self.returns_with_name_regex = re.compile(r"(\w+)\s*:\s*([\w\[\], ]+)")

    def get_exceptions_raised(self, doc: str) -> List[str]:
        """Returns the exceptions listed as raised in the docstring.

        Args:
            doc (str): The docstring to analyze.

        Raises:
            ParseException: Raised if any error happens during parsing.

        Returns:
            List[str]: List of exceptions raised.
        """
        if 'Raises' not in doc:
            return []
        try:
            splits = self.section_regex.split(doc)
            raises_idx = splits.index('Raises')
            if raises_idx == -1 or len(splits) < raises_idx + 1:
                raise ParseException()

            exception_section = splits[raises_idx + 1]

            exceptions_raised: List[str] = []
            for exn_line in exception_section.split('\n'):
                # Heuristic: Assume errors-types are not indented and descriptions are.
                if not exn_line.startswith(' '):
                    exceptions_raised.append(exn_line)

            return exceptions_raised
        except Exception:
            raise ParseException()

    def get_summary(self, doc: str, module_type: ModuleType) -> Optional[str]:
        """Returns the summary part of the docstring.

        Args:
            doc (str): The docstring to analyze.
            module_type (ModuleType): The module it was extracted from.

        Raises:
            ParseException: Raised if any error happens during parsing.

        Returns:
            Optional[str]: The summary, if it exists.
        """
        if len(doc) == 0:
            return None

        try:
            splits = self.section_regex.split(doc)

            if len(splits) > 0 and splits[0] not in self.section_headers:
                return splits[0]

            return None
        except Exception:
            raise ParseException()

    def get_parameters(self, doc: str, module_type: ModuleType) -> List[Parameter]:
        """Finds the function arguments as strings, and returns their types as Parameter instances.

        Args:
            doc (str): Function docstring.
            module_type (ModuleType): The module the docstring was extracted from.

        Raises:
            ParseException: Raised if any error happens during parsing.

        Returns:
            List[Parameter]: The parameters parsed from the docstring.
        """
        if 'Parameters' not in doc:
            return []
        try:
            splits = self.section_regex.split(doc)
            parameters_idx = splits.index('Parameters')
            if parameters_idx == -1 or len(splits) < parameters_idx + 1:
                raise ParseException()

            parameters_section = splits[parameters_idx + 1]
            matches = self.parameter_regex.findall(parameters_section)

            # If we have a parameters section with text, but no matches, we must have bad formatting
            if len(parameters_section) > 0 and len(matches) == 0:
                raise ParseException()

            parameters: List[Parameter] = []
            for param_name, param_type in matches:
                located_type = get_type_from_module(param_type, module_type)
                parameters.append(Parameter(param_name, located_type.type))
            return parameters
        except Exception:
            raise ParseException()

    def get_return_type(self, doc: str, module_type: ModuleType) -> Type:
        """Base method for Parsers to return the return-type of a function from its docstring.

        Args:
            doc (str): Function docstring.
            module_type (ModuleType): The module the docstring was extracted from.

        Raises:
            ParseException: Raised if an exception occurs during parsing

        Returns:
            Type: The return type parsed from the docs.
        """
        # TODO: Use section split regex so we know 'Returns' is not just a random word in docstring, but actually a section header
        if 'Returns' not in doc:
            return type(None)

        try:
            splits = self.section_regex.split(doc)
            parameters_idx = splits.index('Returns')
            if parameters_idx == -1 or len(splits) < parameters_idx + 1:
                raise ParseException()

            returns_section = splits[parameters_idx + 1]
            returns_line = returns_section.split('\n')[0]

            # Test for line with name/type combo
            if ':' in returns_line:
                match = self.returns_with_name_regex.match(returns_line)
                if match is not None:
                    name, doctype = match.groups()
                else:
                    raise ParseException()
            else:
                doctype = returns_line.rstrip('\n')

            return get_type_from_module(doctype, module_type).type
        except Exception:
            raise ParseException()



from pydoctest.configuration import Configuration
from pydoctest.validation import ResultType, validate_class, validate_function
from pydoctest.main import PyDoctestService

import tests.test_class.correct_class
import tests.test_class.nodocstring_class
import tests.test_class.incorrect_class
import tests.test_class.raises_class


class TestDocs():
    # test correct_class
    def test_correct_class_module(self) -> None:
        config = Configuration.get_configuration_from_path("tests/test_class/pydoctest_correct_class.json")

        ds = PyDoctestService(config)
        result = ds.validate()
        assert result.result == ResultType.OK

    def test_correct_class_class_is_ok(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_class(tests.test_class.correct_class.CorrectTestClass, config, tests.test_class.correct_class)
        assert result.result == ResultType.OK
        for res in result.function_results:
            assert res.result == ResultType.OK

    def test_correct_class_empty_func(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.correct_class.CorrectTestClass.empty_func, config, tests.test_class.correct_class)
        assert result.result == ResultType.OK

    def test_correct_class_func_returns_none(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.correct_class.CorrectTestClass.func_returns_none, config, tests.test_class.correct_class)
        assert result.result == ResultType.OK

    def test_correct_class_returns_int(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.correct_class.CorrectTestClass.func_returns_int, config, tests.test_class.correct_class)
        assert result.result == ResultType.OK

    def test_correct_class_has_arg_returns_arg(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.correct_class.CorrectTestClass.func_has_arg_returns_arg, config, tests.test_class.correct_class)
        assert result.result == ResultType.OK

    # Test nodocstring_class
    def test_func_no_docstring_no_fail(self) -> None:
        config = Configuration.get_default_configuration()
        config.fail_on_missing_docstring = False
        result = validate_function(tests.test_class.nodocstring_class.ClassNoDocString.func_no_docstring, config, tests.test_class.correct_class)
        assert result.result == ResultType.NO_DOC

    def test_func_no_docstring_do_fail(self) -> None:
        config = Configuration.get_default_configuration()
        config.fail_on_missing_docstring = True
        result = validate_function(tests.test_class.nodocstring_class.ClassNoDocString.func_no_docstring, config, tests.test_class.correct_class)
        assert result.result == ResultType.FAILED

    def test_func_no_summary_no_fail(self) -> None:
        config = Configuration.get_default_configuration()
        config.fail_on_missing_summary = False
        result = validate_function(tests.test_class.nodocstring_class.ClassNoDocString.func_no_summary, config, tests.test_class.correct_class)
        assert result.result == ResultType.OK

    def test_func_no_summary_do_fail(self) -> None:
        config = Configuration.get_default_configuration()
        config.fail_on_missing_summary = True
        result = validate_function(tests.test_class.nodocstring_class.ClassNoDocString.func_no_summary, config, tests.test_class.correct_class)
        assert result.result == ResultType.FAILED
        assert result.fail_reason == 'Function does not have a summary'

    def test_func_with_raises(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.raises_class.RaisesClass.func_with_raise_and_args_and_return, config, tests.test_class.incorrect_class)
        assert result.result == ResultType.OK

    # Test incorrect_class
    def test_incorrect_class_module(self) -> None:
        config = Configuration.get_configuration_from_path("tests/test_class/pydoctest_incorrect_class.json")
        ds = PyDoctestService(config)
        result = ds.validate()
        assert result.result == ResultType.FAILED

    def test_incorrect_class_is_fail(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_class(tests.test_class.incorrect_class.IncorrectTestClass, config, tests.test_class.incorrect_class)
        assert result.result == ResultType.FAILED
        for res in result.function_results:
            assert res.result == ResultType.FAILED

    def test_incorrect_class_empty_func(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.incorrect_class.IncorrectTestClass.empty_func, config, tests.test_class.incorrect_class)
        assert result.result == ResultType.FAILED
        assert result.fail_reason == "Number of arguments differ. Expected (from signature) 0 arguments, but found (in docs) 1."

    def test_incorrect_class_func_returns_none(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.incorrect_class.IncorrectTestClass.func_returns_none, config, tests.test_class.incorrect_class)
        assert result.result == ResultType.FAILED
        assert result.fail_reason == "Return type differ. Expected (from signature) <class 'NoneType'>, but got (in docs) <class 'int'>."

    def test_incorrect_class_func_returns_int(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.incorrect_class.IncorrectTestClass.func_returns_int, config, tests.test_class.incorrect_class)
        assert result.result == ResultType.FAILED
        assert result.fail_reason == "Return type differ. Expected (from signature) <class 'int'>, but got (in docs) <class 'bool'>."

    def test_incorrect_class_func_has_arg_returns_arg(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.incorrect_class.IncorrectTestClass.func_has_arg_returns_arg, config, tests.test_class.incorrect_class)
        assert result.result == ResultType.FAILED
        assert result.fail_reason == "Return type differ. Expected (from signature) <class 'float'>, but got (in docs) <class 'NoneType'>."

    def test_incorrect_class_func_name_mismatch(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.incorrect_class.IncorrectTestClass.func_name_mismatch, config, tests.test_class.incorrect_class)
        assert result.result == ResultType.FAILED
        assert result.fail_reason == "Argument name differ. Expected (from signature) 'a', but got (in docs) 'b'"

    def test_incorrect_class_func_type_mismatch(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.incorrect_class.IncorrectTestClass.func_type_mismatch, config, tests.test_class.incorrect_class)
        assert result.result == ResultType.FAILED
        assert result.fail_reason == "Argument type differ. Argument 'a' was expected (from signature) to have type '<class 'int'>', but has (in docs) type '<class 'float'>'"

    # Test counts_class
    def test_counts_class(self) -> None:
        config = Configuration.get_configuration_from_path("tests/test_class/pydoctest_get_counts.json")
        config.fail_on_missing_docstring = False

        ds = PyDoctestService(config)
        result = ds.validate()
        counts = result.get_counts()

        assert counts.functions_failed == 1
        assert counts.functions_skipped == 1
        assert counts.functions_succeeded == 2
        assert counts.module_count == 1

        assert counts.get_total() == 4

    # Test enum_in_module
    def test_enum_in_module(self) -> None:
        # Test that enums are ignored
        config = Configuration.get_configuration_from_path("tests/test_class/pydoctest_enum_in_module.json")
        ds = PyDoctestService(config)
        result = ds.validate()
        assert len(result.module_results[0].class_results) == 1
        assert result.module_results[0].class_results[0].class_name == "ExampleClass"

    # Test no include_paths in config
    def test_no_include_paths(self) -> None:
        config = Configuration.get_configuration_from_path("tests/test_class/test_no_include_paths/pydoctest.json")
        ds = PyDoctestService(config)
        result = ds.validate()
        assert result.result == ResultType.OK

        # Assert that 1 module was found in root. We do not search recursively.
        assert len(result.module_results) == 1

    # Test incorrect raise
    def test_func_with_missing_raise(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.raises_class.RaisesClass.func_with_missing_raise, config, tests.test_class.incorrect_class)
        assert result.result == ResultType.FAILED

    def test_func_with_incorrect_raise(self) -> None:
        config = Configuration.get_default_configuration()
        result = validate_function(tests.test_class.raises_class.RaisesClass.func_with_incorrect_raise, config, tests.test_class.incorrect_class)
        assert result.result == ResultType.FAILED

    def test_fail_on_raises_section_dont_fail(self) -> None:
        config = Configuration.get_default_configuration()
        config.fail_on_raises_section = False
        result = validate_function(tests.test_class.raises_class.RaisesClass.func_with_incorrect_raise, config, tests.test_class.incorrect_class)
        assert result.result == ResultType.OK


import subprocess
import sys

from typing import Tuple
from pydoctest.configuration import Configuration
from pydoctest.main import PyDoctestService
from enum import Enum


class OS(Enum):
    WINDOWS = 0,
    LINUX = 1


def get_os() -> OS:
    if sys.platform == 'win32':
        return OS.WINDOWS
    else:
        return OS.LINUX


class TestPydoctest():
    """
    Uses the library to test its own docstrings - how awesome is that?
    """
    def test_pydoctest(self) -> None:
        config = Configuration.get_configuration_from_path("pydoctest.json")
        ds = PyDoctestService(config)
        result = ds.validate()
        counts = result.get_counts()
        assert counts.functions_failed == 0, "Run pydoctest on project and fix docstrings."
        # TODO: This doesn't artificially bring up our code-coverage, right?


class TestCase():
    def execute_command(self, command: str) -> Tuple[str, str]:
        try:
            if get_os() == OS.WINDOWS:
                # Figure out a more robust way to handle OS
                if command.startswith('python3 '):
                    command = command.replace('python3', 'py', 1)
                p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            else:
                p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env={'LC_ALL': 'C'})
            p.wait()
            stdout, stderr = p.communicate()
            return str(stdout.decode("utf-8")), str(stderr.decode("utf-8"))
        except Exception as e:
            return "", str(e)



class ClassNoDocString():
    def func_with_docstring(self, a: int) -> int:
        """[summary]

        Args:
            a (int): [description]

        Returns:
            int: [description]
        """
        pass

    def func_no_docstring(self, a: int) -> int:
        pass

    def func_no_summary(self, a: int) -> int:
        """

        Args:
            a (int): [description]

        Returns:
            int: [description]
        """
        pass



import os
import sys
import json
from typing import Any, Dict, List, Optional
from enum import IntEnum

from pydoctest.logging import log
from pydoctest.parsers.parser import Parser
from pydoctest.parsers.google_parser import GoogleParser
from pydoctest.parsers.numpy_parser import NumpyParser
from pydoctest.parsers.sphinx_parser import SphinxParser


class Verbosity(IntEnum):
    QUIET = 0
    SHOW_FAILED = 1
    SHOW_ALL = 2


PARSERS = {
    'google': GoogleParser,
    'numpy': NumpyParser,
    'sphinx': SphinxParser
}


class Configuration():
    def __init__(self) -> None:
        """Configuration object to be used while running pydoctest.
        The properties here can be overridden with a pydoctest.json config file.
        """
        self.working_directory: str = ""

        # List of patterns to discover modules by
        self.include_paths: List[str] = [ '**/*.py' ]

        # List of patterns to exclude modules discovered by
        self.exclude_paths: List[str] = [ '**/__init__.py', '**/setup.py' ]

        # Doctype parser to use, defaults to Google
        self.parser = "google"

        # Verbosity of reporter, currently only used by text-reporter
        self.verbosity = Verbosity.SHOW_FAILED

        # Throw an error if function does not have a docstring
        self.fail_on_missing_docstring = False

        # Throw an error if function does not have a summary
        self.fail_on_missing_summary = False

        # Throw an error if 'raises' section does not list all exceptions
        self.fail_on_raises_section = True

    @staticmethod
    def get_default_configuration(root_dir: Optional[str] = None) -> 'Configuration':
        """Returns a configuration with default values.

        Args:
            root_dir (Optional[str]): Directory to use as root.

        Returns:
            'Configuration': A default configuration.
        """
        log("Using default configuration")
        config = Configuration()
        if root_dir:
            config.working_directory = root_dir
        return config

    @staticmethod
    def get_configuration_from_path(config_path: str) -> 'Configuration':
        """Returns a configuration, loaded from the pydoctest.json provided.

        Args:
            config_path (str): The path to a config file.

        Returns:
            'Configuration': A configuration loaded with values from provided path.
        """
        config_path_absolute = os.path.abspath(config_path)
        log(f"Using configuration from path: {config_path_absolute}")

        with open(config_path_absolute, 'r') as f:
            config_dict = json.load(f)
            config = Configuration.from_dict(config_dict)
            config.working_directory = os.path.dirname(config_path_absolute)
            return config

    @staticmethod
    def from_dict(x: Dict[str, Any]) -> 'Configuration':
        """Given a dictionary, returns a Configuration object.

        Args:
            x (Dict[str, Any]): The dictionary to load the configuration values from.

        Returns:
            'Configuration': A configuration loaded with values from the dictionary.
        """
        obj = Configuration()
        for key, value in x.items():
            if hasattr(obj, key):
                setattr(obj, key, value)
            else:
                raise Exception(f'Unknown configuration parameter: {key}')
        return obj

    def get_parser(self) -> Parser:
        """Checks if the desired Parser exists and returns it.

        Raises:
            Exception: If parser from Configuration doesn't exist.

        Returns:
            Parser: A supported Parser.
        """
        if self.parser in PARSERS.keys():
            return PARSERS[self.parser]()
        else:
            raise Exception(f"Unknown parser: {self.parser}. Please use one of the following: {', '.join(PARSERS.keys())}")



import pydoc
from typing import Any, Dict
import pytest

from pydoctest.parsers.numpy_parser import NumpyParser
from pydoctest.exceptions import ParseException

import tests.test_parsers.numpy_class


class TestNumpyParser():

    def test_parse_exception_get_parameters(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.IncorrectTestClass.func_parse_exception)
        with pytest.raises(ParseException) as exc_info:
            parser.get_parameters(doc, tests.test_parsers.numpy_class)

    def test_parse_exception_get_return_type(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.IncorrectTestClass.func_parse_exception)
        with pytest.raises(ParseException) as exc_info:
            parser.get_return_type(doc, tests.test_parsers.numpy_class)

    def test_empty_func(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.CorrectTestClass.empty_func)
        arguments = parser.get_parameters(doc, tests.test_parsers.numpy_class)
        assert len(arguments) == 0, f"NumpyParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.numpy_class)
        assert return_type == type(None), f"NumpyParser failed assertion"

    def test_func_returns_none(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.CorrectTestClass.func_returns_none)
        arguments = parser.get_parameters(doc, tests.test_parsers.numpy_class)
        assert len(arguments) == 1, f"NumpyParser failed assertion"
        assert arguments[0].type == int, f"NumpyParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.numpy_class)
        assert return_type == type(None), f"NumpyParser failed assertion"

    def test_func_returns_int(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.CorrectTestClass.func_returns_int)
        arguments = parser.get_parameters(doc, tests.test_parsers.numpy_class)
        assert len(arguments) == 0, f"NumpyParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.numpy_class)
        assert return_type == int, f"NumpyParser failed assertion"

    def test_func_returns_int_name_type(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.CorrectTestClass.func_returns_int_name_type)
        arguments = parser.get_parameters(doc, tests.test_parsers.numpy_class)
        assert len(arguments) == 0, f"NumpyParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.numpy_class)
        assert return_type == int, f"NumpyParser failed assertion"

    def test_func_has_arg_returns_arg(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.CorrectTestClass.func_has_arg_returns_arg)
        arguments = parser.get_parameters(doc, tests.test_parsers.numpy_class)
        assert len(arguments) == 1, f"NumpyParser failed assertion"
        assert arguments[0].type == int, f"NumpyParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.numpy_class)
        assert return_type == float, f"NumpyParser failed assertion"

    def test_func_has_raises_doc(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.CorrectTestClass.func_has_raises_doc)
        arguments = parser.get_parameters(doc, tests.test_parsers.numpy_class)
        assert len(arguments) == 1, f"NumpyParser failed assertion"
        assert arguments[0].type == int, f"NumpyParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.numpy_class)
        assert return_type == int, f"NumpyParser failed assertion"

    def test_func_with_multiline_summary(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.CorrectTestClass.func_with_multiline_summary)
        arguments = parser.get_parameters(doc, tests.test_parsers.numpy_class)
        assert len(arguments) == 1, f"NumpyParser failed assertion"
        assert arguments[0].type == int, f"NumpyParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.numpy_class)
        assert return_type == int, f"NumpyParser failed assertion"

    def test_get_summary_multiline_summary(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.CorrectTestClass.func_with_multiline_summary)

        summary = parser.get_summary(doc, tests.test_parsers.numpy_class)
        assert summary is not None
        assert len(summary) > 0, f"NumpyParser failed assertion"
        assert(len([x for x in summary if x == '\n']) > 1), f"NumpyParser failed assertion"

    def test_get_summary_empty_summary(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.CorrectTestClass.func_no_summary)
        arguments = parser.get_parameters(doc, tests.test_parsers.numpy_class)
        assert len(arguments) == 0, f"NumpyParser failed assertion"

        return_type = parser.get_return_type(doc, tests.test_parsers.numpy_class)
        assert return_type == type(None), f"NumpyParser failed assertion"

        summary = parser.get_summary(doc, tests.test_parsers.numpy_class)
        assert summary is None, f"NumpyParser failed assertion"

    def test_func_with_raise_and_args_and_return(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.RaisesClass.func_with_raise_and_args_and_return)
        actual_exceptions = parser.get_exceptions_raised(doc)
        expected_exceptions = [ 'RuntimeError', 'ValueError', 'IndexError' ]
        assert len(expected_exceptions) == len(actual_exceptions)

        intersection = set(expected_exceptions) - set(actual_exceptions)
        assert len(intersection) == 0

    def test_func_with_raise_and_args(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.RaisesClass.func_with_raise_and_args)
        actual_exceptions = parser.get_exceptions_raised(doc)
        expected_exceptions = [ 'RuntimeError', 'ValueError', 'IndexError' ]
        assert len(expected_exceptions) == len(actual_exceptions)

        intersection = set(expected_exceptions) - set(actual_exceptions)
        assert len(intersection) == 0

        parameters = parser.get_parameters(doc, tests.test_parsers.numpy_class)
        assert len(parameters) == 2
        assert parameters[0].name == 'a'
        assert parameters[0].type == int

        assert parameters[1].name == 'b'
        assert parameters[1].type == float

    def test_func_with_raise(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.RaisesClass.func_with_raise)
        actual_exceptions = parser.get_exceptions_raised(doc)
        expected_exceptions = [ 'RuntimeError', 'ValueError', 'IndexError' ]
        assert len(expected_exceptions) == len(actual_exceptions)

        intersection = set(expected_exceptions) - set(actual_exceptions)
        assert len(intersection) == 0

    def test_func_with_generics(self) -> None:
        parser = NumpyParser()
        doc = pydoc.getdoc(tests.test_parsers.numpy_class.CorrectTestClass.func_with_generics)
        parameters = parser.get_parameters(doc, tests.test_parsers.numpy_class)
        assert len(parameters) == 1
        assert parameters[0].type == Dict[str, Any]
        assert parameters[0].name == 'a_a'

        return_type = parser.get_return_type(doc, tests.test_parsers.numpy_class)
        assert return_type == Dict[str, Any]



from enum import Enum
import sys
import os
import argparse
import pathlib
import inspect
import traceback

import importlib
import importlib.util
from importlib.abc import Loader
from importlib.machinery import ModuleSpec

from types import FunctionType, ModuleType
from typing import List, Optional, Type

from pydoctest import logging
from pydoctest.version import VERSION
from pydoctest.configuration import Configuration, Verbosity
from pydoctest.reporters.reporter import Reporter
from pydoctest.reporters.json_reporter import JSONReporter
from pydoctest.reporters.text_reporter import TextReporter
from pydoctest.validation import ModuleValidationResult, Result, ResultType, ValidationResult, validate_class, validate_function
from pydoctest.utilities import parse_cli_list

# We always want to exclude setup.py
DEFAULT_EXCLUDE_PATHS = [ "**/setup.py" ]

CONFIG_FILE_NAME = 'pydoctest.json'
REPORTERS = {
    'json': JSONReporter,
    'text': TextReporter
}


class PyDoctestService():
    def __init__(self, config: Configuration) -> None:
        """Instantiates a new Pydoctest-service.

        Args:
            config (Configuration): The configuration to use for testing.
        """
        self.config = config

    def validate(self, modules: Optional[List[str]] = None) -> ValidationResult:
        """Validate the found modules using the provided reporter.

        Args:
            modules (Optional[List[str]]): Optionally, specify directly the modules rather than discover.

        Returns:
            ValidationResult: Information about whether validation succeeded.
        """
        logging.log('Starting validating')
        result = ValidationResult()

        if modules is None:
            modules = self.discover_modules()
            logging.log(f'Found {len(modules)} modules')

        for module in modules:
            module_result = self.validate_module(module)
            if module_result.result == ResultType.FAILED:
                result.result = ResultType.FAILED
            result.module_results.append(module_result)

        if result.result == ResultType.NOT_RUN:
            result.result = ResultType.OK

        return result

    def validate_module(self, module_path: str) -> ModuleValidationResult:
        """Validates the module, given its path.

        Args:
            module_path (str): Path to a module.

        Returns:
            ModuleValidationResult: Result of validating the module.
        """
        logging.log(f'Validating module: {module_path}')
        result = ModuleValidationResult(module_path)

        module_name = os.path.basename(module_path)
        module_spec: Optional[ModuleSpec] = importlib.util.spec_from_file_location(module_name, module_path)

        if not os.path.exists(module_path) or module_spec is None or not isinstance(module_spec.loader, Loader):
            result.result = ResultType.NOT_RUN
            result.fail_reason = f"Failed to load file from location: {module_path}"
            return result

        try:
            module_type = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(module_type)
        except ModuleNotFoundError as e:
            result.result = ResultType.FAILED
            result.fail_reason = f"Failed to load module dependant module: {str(e)}"
            return result
        except Exception as e:
            result.result = ResultType.FAILED
            result.fail_reason = f"Failed to load module (possibly due to syntax errors): {module_path} - error: {str(e)}"
            return result

        # Validate top-level functions in module
        fns = self.get_global_functions(module_type)
        for fn in fns:
            function_result = validate_function(fn, self.config, module_type)
            if function_result.result == ResultType.FAILED:
                result.result = ResultType.FAILED
            result.function_results.append(function_result)

        # Validate top-level classes in module
        classes = self.get_classes(module_type)
        for cl in classes:
            class_result = validate_class(cl, self.config, module_type)
            if class_result.result == ResultType.FAILED:
                result.result = ResultType.FAILED
            result.class_results.append(class_result)

        return result

    def get_global_functions(self, module: ModuleType) -> List[FunctionType]:
        """Gets the global functions of the module.

        Args:
            module (ModuleType): The module to extract global functions from.

        Returns:
            List[FunctionType]: A list of global functions in the module.
        """
        fns = []
        for name, obj in inspect.getmembers(module, lambda x: inspect.isfunction(x) and x.__module__ == module.__name__):
            fns.append(obj)
        return fns

    def get_classes(self, module: ModuleType) -> List[Type]:
        """Get classes defined in module.

        Args:
            module (ModuleType): The module to extract classes from.

        Returns:
            List[Type]: A list of classes defined in the module.
        """
        classes: List[type] = []
        # getmembers will also return imports etc. so we require classes to be part of this module
        for name, obj in inspect.getmembers(module, lambda x: inspect.isclass(x) and x.__module__ == module.__name__):
            if issubclass(obj, Enum):
                # Ignore enums
                continue
            classes.append(obj)
        return classes

    def __is_excluded_path(self, path: pathlib.Path, exclude_paths: List[str]) -> bool:
        """
        Returns whether the found path is excluded by any of the exclude_paths.

        Args:
            path (pathlib.Path): The path to test
            exclude_paths (List[str]): The exclude paths

        Returns:
            bool: If path is excluded.
        """
        return any(path.match(e_p) for e_p in exclude_paths)

    def discover_modules(self) -> List[str]:
        """Discovers modules using the configuration include/exclude paths.

        Returns:
            List[str]: A list of paths to modules to be validated.
        """
        include_file_paths: List[str] = []

        include_paths = self.config.include_paths
        exclude_paths = self.config.exclude_paths + DEFAULT_EXCLUDE_PATHS

        for include_path in include_paths:
            path = pathlib.Path(self.config.working_directory)
            disovered_paths = list(path.glob(include_path))
            allowed_paths = [str(p) for p in disovered_paths if not self.__is_excluded_path(p, exclude_paths)]
            include_file_paths.extend(allowed_paths)

        return include_file_paths


def get_configuration(root_dir: str, config_path: Optional[str] = None) -> Configuration:
    """Searches for CONFIG_FILE_NAME in root_dir, unless a path is provided.

    Args:
        root_dir (str): The directory to search in.
        config_path (Optional[str]): [description]. Defaults to None.

    Returns:
        Configuration: Either a configuration matching the specified/found one, or a default one.
    """
    if config_path:
        return Configuration.get_configuration_from_path(config_path)

    config_paths = [p for p in os.listdir(root_dir) if p == CONFIG_FILE_NAME]
    if len(config_paths) == 0:
        # TODO: Is the interface better by returning Optional[Configuration] (None here)?
        # Then it is clearer that a config was not found.
        return Configuration.get_default_configuration(root_dir)

    path = os.path.join(root_dir, config_paths[0])
    return Configuration.get_configuration_from_path(path)


def get_reporter(config: Configuration, reporter: Optional[str] = None) -> Reporter:
    """We offer to output results using either TextReporter and JSONReporter.
    This list of reporters can be extended with more reporters as they simply implement a get_output function.

    Args:
        config (Configuration): The configuration currently used.
        reporter (Optional[str]): Desired reporter [text | json]

    Raises:
        Exception: Raised if desired reporter does not exist.

    Returns:
        Reporter: Reporter if provided, otherwise text.
    """
    if reporter is None:
        return REPORTERS['text'](config)

    if reporter in REPORTERS.keys():
        return REPORTERS[reporter](config)
    else:
        raise Exception(f"Unknown reporter: {reporter}. Please use one of the following: {', '.join(REPORTERS.keys())}")


def main() -> None:  # pragma: no cover
    """Main function invoked when running script.
    """
    # TODO: Could allow arguments directly to pydoctest for overriding .json config arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", help="Path to config JSON file, e.g. pydoctest.json")
    parser.add_argument("--reporter", help="Reporter to use, either 'json' or 'text'")
    parser.add_argument("--verbosity", help="0 = quiet, 1 = show failed, 2 = show all")
    parser.add_argument("--debug", help="Verbose logging", action='store_true')
    parser.add_argument("--version", help="Show version", action='store_true')
    parser.add_argument("--file", help="Analyze single file")
    parser.add_argument("--parser", help="Docstring format, either: google|sphinx|numpy")

    # TODO: Implement splitting by comma and
    parser.add_argument("--include-paths", help="Pattern to include paths by, defaults to \"**/*.py\"")
    parser.add_argument("--exclude-paths", help="Pattern to exclude paths by, defaults to \"**/__init__.py, **/setup.py\"")

    args = parser.parse_args()

    try:
        if args.version:
            print(VERSION)
            sys.exit(0)

        if args.debug:
            logging.set_verbose(True)

        config = get_configuration(os.getcwd(), args.config)

        # Imports will not work, unless we pretend this script is executed in the current directory.
        sys.path.insert(0, '')

        reporter = get_reporter(config, args.reporter)

        if args.verbosity:
            config.verbosity = Verbosity(int(args.verbosity))

        if args.parser:
            config.parser = args.parser

        if args.include_paths:
            config.include_paths = parse_cli_list(args.include_paths)

        if args.exclude_paths:
            config.exclude_paths = parse_cli_list(args.exclude_paths)

        # Check that parser exists before running.
        config.get_parser()

        ds = PyDoctestService(config)

        if args.file:
            result = ds.validate([os.path.abspath(args.file)])
        else:
            result = ds.validate()

        output = reporter.get_output(result)

        if isinstance(reporter, TextReporter) and config.verbosity != Verbosity.QUIET:
            counts = result.get_counts()
            output += f"Tested {counts.get_total()} function(s) across {counts.module_count} module(s).\n"
            output += f"Succeeded: {counts.functions_succeeded}, Failed: {counts.functions_failed}, Skipped: {counts.functions_skipped}"

        print(output)

        if result.result != ResultType.OK:
            sys.exit(1)
    except Exception as e:
        print(traceback.format_exc())
        print(f"Error occurred: {str(e)}")
        sys.exit(1)


if __name__ == '__main__':
    main()



class GoogleClass():
    def function_with_colon(self, a: int, b: int) -> None:
        """Summary

            Args:
                a (int): here's a string with a colon :
                b (int): A normal parameter
        """
        pass



from collections import deque
import inspect
import ast
import textwrap

from types import FunctionType, ModuleType
from typing import List, Type, cast

from pydoc import locate

from pydoctest.exceptions import UnknownTypeException


class LocateResult():
    """Small DTO for storing type and method used for finding it.
    Makes testing easier.
    """
    def __init__(self, t: Type, method: str) -> None:
        """Instantiates a new LocateResult

        Args:
            t (Type): The type returned.
            method (str): The method used to locate it.
        """
        self.type = t
        self.method = method


def get_type_from_module(type_string: str, module: ModuleType) -> LocateResult:
    """Attempts to return the type, given the type_string and module it is extracted from.

    Args:
        type_string (str): String version of a string, e.g. "Optional[str]"
        module (ModuleType): The module the type_string is extracted from.

    Raises:
        UnknownTypeException: If unable to find the type, we throw an Exception.

    Returns:
        LocateResult: A LocateResult wrapping the type when found.
    """
    # First let pydoc attempt to locate the type
    located_type: Type = cast(Type, locate(type_string))
    if located_type:
        return LocateResult(located_type, 'locate')

    # Try to eval it.
    try:
        # We pass the globals of module to eval, so lookups should work.
        t = eval(type_string, vars(module))
        return LocateResult(t, 'eval')
    except NameError:
        pass

    # Search imported modules in module
    # Inspired by this: https://stackoverflow.com/a/11781721/3717691
    # TODO: We limit the search by count, not by depth. Should it be configurable, e.g. when large codebases?
    SEARCHES_LEFT = 1_000
    q = deque([module])
    while q and SEARCHES_LEFT > 0:
        m = q.popleft()
        try:
            t = eval(type_string, vars(m))
            return LocateResult(t, 'deque')
        except NameError:
            pass

        for name, item in inspect.getmembers(m):
            if inspect.ismodule(item):
                q.append(item)
            elif inspect.isclass(item) and item.__module__ != module.__name__:
                mod = inspect.getmodule(item)
                if mod:
                    q.append(mod)
        SEARCHES_LEFT -= 1

    # TODO: We should make this configurable in config
    raise UnknownTypeException(f"Was unable to detect the type of: {type_string} from module: {module.__file__}.\nIf you believe this is a bug, please file it here: https://github.com/jepperaskdk/pydoctest/issues")


class RaiseVisitor(ast.NodeVisitor):
    def __init__(self) -> None:
        """A NodeVisitor which collects the id's of ast.Raise objects from parsing the function.
        """
        self.nodes: List[str] = []

    def visit(self, n: ast.AST) -> None:
        """Visits the node and if the node is an ast.Raise, we collect its id.

        Args:
            n (ast.AST): The node to test.

        """
        if isinstance(n, ast.Raise) and n.exc and isinstance(n.exc, ast.Call) and isinstance(n.exc.func, ast.Name):
            self.nodes.append(n.exc.func.id)
        super().visit(n)


def get_exceptions_raised(fn: FunctionType, module: ModuleType) -> List[str]:
    """Get exceptions raised in fn.

    NOTE: We currently only return exceptions explicitly raised by the function.
    This means, that exceptions raised from super-calls are not returned,
    and exceptions thrown by called functions in the function are not returned.

    Args:
        fn (FunctionType): The function to get raised exceptions from.
        module (ModuleType): The module in which the function is defined.

    Returns:
        List[str]: The list of exceptions thrown.
    """
    # TODO: How do we compile functions, without having to check indentation

    fn_source = inspect.getsource(fn)
    max_indents = 10

    # Try to parse, and if IndentationError, dedent up to 10 times.
    while max_indents > 0:
        try:
            tree = ast.parse(fn_source)
            break
        except IndentationError as e:
            fn_source = textwrap.dedent(fn_source)
            max_indents -= 1

    visitor = RaiseVisitor()
    visitor.generic_visit(tree)

    # Make sure we dont return duplicate exceptions
    return list(set(visitor.nodes))


def parse_cli_list(content: str, separator: str = ',') -> List[str]:
    """
    Parses a string-list by splitting on separator, trimming and removing empty results.

    Args:
        content (str): The string coming from a cli command.
        separator (str): The separator to split the list by. Defaults to ','.

    Returns:
        List[str]: The list-items.
    """
    items = content.split(separator)
    return [i.strip() for i in items if i]



import re

from types import ModuleType
from typing import List, Optional, Type

from pydoctest.parsers.parser import Parameter, Parser
from pydoctest.utilities import get_type_from_module
from pydoctest.exceptions import ParseException

"""
Based on this style guide:
https://sphinx-rtd-tutorial.readthedocs.io/en/latest/docstrings.html

"""


class SphinxParser(Parser):
    def __init__(self) -> None:
        """Parser for Sphinx docstring style."""
        super().__init__()
        self.parameter_name_regex = re.compile(r":param\s+(\w+):")
        self.parameter_type_regex = re.compile(r":type\s+\w+:\s*([\w\[\], ]+)")
        self.return_type_regex = re.compile(r":rtype:\s*([\w\[\], ]+)")
        self.raises_regex = re.compile(r":raises\s+(\w+):")

    def get_exceptions_raised(self, doc: str) -> List[str]:
        """Returns the exceptions listed as raised in the docstring.

        Args:
            doc (str): The docstring to analyze.

        Raises:
            ParseException: Raised if any error happens during parsing.

        Returns:
            List[str]: List of exceptions raised.
        """
        try:
            raised_exceptions: List[str] = []
            for line in doc.split('\n'):
                match = self.raises_regex.match(line)
                if match is None:
                    continue
                doctype = match.groups()[0]
                raised_exceptions.append(doctype)
            return raised_exceptions
        except Exception:
            raise ParseException()

    def get_summary(self, doc: str, module_type: ModuleType) -> Optional[str]:
        """Returns the summary part of the docstring.

        Args:
            doc (str): The docstring to analyze.
            module_type (ModuleType): The module it was extracted from.

        Raises:
            ParseException: Raised if any error happens during parsing.

        Returns:
            Optional[str]: The summary, if it exists.
        """
        try:
            # Find first occurance of :param, :type, :return, :rtype, :raises
            indices = [doc.find(x) for x in [':param', ':type', ':return', ':rtype', ':raises'] if doc.find(x) >= 0]
            if len(indices) == 0:
                return doc if len(doc) > 0 else None

            min_index = min(indices)
            summary = doc[0: min_index].rstrip()
            return summary if len(summary) > 0 else None
        except Exception as e:
            raise ParseException()

    def get_parameters(self, doc: str, module_type: ModuleType) -> List[Parameter]:
        """Finds the function arguments as strings, and returns their types as Parameter instances.

        Args:
            doc (str): Function docstring.
            module_type (ModuleType): The module the docstring was extracted from.

        Raises:
            ParseException: Raised if any error happens during parsing.

        Returns:
            List[Parameter]: The parameters parsed from the docstring.
        """
        try:
            lines = doc.split('\n')
            parameters: List[Parameter] = []
            var_name: Optional[str] = None
            for line in lines:
                if var_name is None:
                    match = self.parameter_name_regex.match(line)
                    if match is not None:
                        var_name = match.groups()[0]
                else:
                    match = self.parameter_type_regex.match(line)
                    if match is not None:
                        var_type = match.groups()[0]
                        located_type = get_type_from_module(var_type, module_type)
                        parameters.append(Parameter(var_name, located_type.type))
                        var_name = None
            if var_name is not None:
                # There must have been an error parsing :type:
                raise ParseException()

            return parameters
        except Exception:
            raise ParseException()

    def get_return_type(self, doc: str, module_type: ModuleType) -> Type:
        """Base method for Parsers to return the return-type of a function from its docstring.

        Args:
            doc (str): Function docstring.
            module_type (ModuleType): The module the docstring was extracted from.

        Raises:
            ParseException: Raised if an exception occurs during parsing

        Returns:
            Type: The return type parsed from the docs.
        """
        try:
            for line in doc.split('\n'):
                match = self.return_type_regex.match(line)
                if match is None:
                    continue
                doctype = match.groups()[0]
                return get_type_from_module(doctype, module_type).type
            return type(None)
        except Exception:
            raise ParseException()



import re
from types import ModuleType
from typing import Dict, List, Optional, Type

from pydoctest.parsers.parser import Parameter, Parser, Section
from pydoctest.utilities import get_type_from_module
from pydoctest.exceptions import ParseException, UnknownTypeException


SECTION_NAMES = {
    Section.ARGUMENTS: "Args:",
    Section.RAISES: "Raises:",
    Section.RETURNS: "Returns:",
}

# Regex matches [name] ([type]): [description] with some extra whitespace
# e.g. this would also match: a      (  int   )   :    kmdkfmdf
# It terminates with .* meaning a new match is the terminator. This should support multiline descriptions without having to consider tabs/indentation.
ARGUMENT_REGEX = re.compile(r"\s*(?P<name>(\w+))\s*\((?P<type>[\w\.\[\], ]+)\)\s*:(.*)")


class GoogleParser(Parser):

    def __get_sections(self, doc: str) -> Dict[Section, str]:
        """
        Splits the docstring into sections.

        Args:
            doc (str): The docstring.

        Returns:
            Dict[Section, str]: A map from section-type to section-string.
        """
        result: Dict[Section, str] = {}
        indices = [(section, doc.index(name)) for section, name in SECTION_NAMES.items() if name in doc]
        sections_sorted = sorted(indices, key=lambda kv: kv[1])

        if len(sections_sorted) > 0:
            result[Section.SUMMARY] = doc[:sections_sorted[0][1]]
        else:
            result[Section.SUMMARY] = doc

        for i in range(len(sections_sorted)):
            section, index = sections_sorted[i]

            if i == len(sections_sorted) - 1:
                # Last section
                result[section] = doc[index:]
            else:
                # Has section after, so use following section's index as stop index
                result[section] = doc[index:sections_sorted[i + 1][1]]

        # Make sure we didn't lose any characters
        assert sum(len(s) for s in result.values()) == len(doc)

        return result

    def get_exceptions_raised(self, doc: str) -> List[str]:
        """Returns the exceptions listed as raised in the docstring.

        Args:
            doc (str): The docstring to analyze.

        Raises:
            ParseException: If unable to parse an argument-line.

        Returns:
            List[str]: List of exceptions raised.
        """
        if 'Raises:' not in doc:
            return []

        try:
            _, tail = doc.split('Raises:')

            if 'Returns:' in tail:
                exceptions, returns = tail.split('Returns:')
            else:
                exceptions = tail

            exceptions_raised: List[str] = []
            for exc_line in [x.strip() for x in exceptions.split("\n") if x]:
                if ':' in exc_line:
                    exceptions_raised.append(exc_line.split(":")[0])
                else:
                    raise ParseException(exc_line)

            return exceptions_raised
        except Exception:
            raise ParseException()

    def get_summary(self, doc: str, module_type: ModuleType) -> Optional[str]:
        """Returns the summary part of the docstring.

        Args:
            doc (str): The docstring to analyze.
            module_type (ModuleType): The module it was extracted from.

        Raises:
            ParseException: If unable to parse an argument-line.

        Returns:
            Optional[str]: The summary, if it exists.
        """
        try:
            for needle in [ 'Args:', 'Raises:', 'Returns:' ]:
                if needle in doc:
                    summ, _ = doc.split(needle)
                    summary = summ.strip()
                    return summary if len(summary) > 0 else None

            if len(doc) > 0:
                return doc.strip()

            return None

        except Exception:
            raise ParseException()

    def get_parameters(self, doc: str, module_type: ModuleType) -> List[Parameter]:
        """Finds the function arguments as strings, and returns their types as Parameter instances.

        Args:
            doc (str): Function docstring.
            module_type (ModuleType): The module the docstring was extracted from.

        Raises:
            ParseException: If unable to parse an argument-line.

        Returns:
            List[Parameter]: The parameters parsed from the docstring.
        """

        sections = self.__get_sections(doc)
        if Section.ARGUMENTS not in sections:
            return []

        parameters = []

        for match in re.finditer(ARGUMENT_REGEX, sections[Section.ARGUMENTS]):
            start, end = match.span()
            name = match.group('name').strip()
            type = match.group('type').strip()

            try:
                located_type = get_type_from_module(type, module_type)
                parameters.append(Parameter(name, located_type.type))
            except UnknownTypeException:
                raise ParseException(f"Unknown type '{type}' in '{sections[Section.ARGUMENTS][start:end].strip()}'")
            except Exception:
                raise ParseException(sections[Section.ARGUMENTS][start:end].strip())

        # If we have an Args section, but no parameters, we must have failed to parse
        if len(parameters) == 0:
            raise ParseException()

        return parameters

    def get_return_type(self, doc: str, module_type: ModuleType) -> Type:
        """Finds the return-type as string and returns it.

        Args:
            doc (str): Function docstring.
            module_type (ModuleType): The module the docstring was extracted from.

        Raises:
            ParseException: If unable to parse the return type line.

        Returns:
            Type: The return type parsed from the docs.
        """
        try:
            if 'Returns:' not in doc:
                return type(None)

            _, returns = doc.split("Returns:")
            try:
                doctype, _ = returns.strip().split(":")
            except ValueError:
                raise ParseException(returns)

            return get_type_from_module(doctype, module_type).type
        except Exception:
            raise ParseException()



from pydoctest.validation import ValidationResult
from pydoctest.configuration import Configuration


class Reporter():
    def __init__(self, config: Configuration) -> None:
        """Creates a new Reporter, which uses the config provided.

        Args:
            config (Configuration): Config which specifies rules about reporting.
        """
        self.config = config

    def get_output(self, result: ValidationResult) -> str:
        """Base function for returning output by walking the ValidationResult object.

        Args:
            result (ValidationResult): The results from running Pydoctest.

        Raises:
            NotImplementedError: Raised if this is not implemented by subclasses.

        Returns:
            str: The output to be returned.
        """
        raise NotImplementedError()



# Module with fixed number of failed, skipped and ok docstrings

def global_func_ok(a: int) -> int:
    """[summary]

    Args:
        a (int): [description]

    Returns:
        int: [description]
    """
    pass


class CountsClass():
    def func_failed(self, b: int) -> int:
        """[summary]

        Args:
            b (float): [description]        <-- should be int

        Returns:
            float: [description]            <-- should be int
        """
        pass

    def func_skipped(self) -> None:
        pass

    def func_ok(self, a: int) -> int:
        """[summary]

        Args:
            a (int): [description]

        Returns:
            int: [description]
        """
        pass
